# -*- coding: utf-8 -*-
from datetime import datetime
import gevent

from flask import (
    Blueprint,
    jsonify,
    Response,
    request,
    make_response,
    copy_current_request_context,
)

from flask_socketio import join_room, leave_room
from flask_login import current_user

DISCONNECT_HANDLED = True


def init_route(app, server, url_prefix):
    bp = Blueprint("remote_access", __name__, url_prefix=url_prefix)

    @bp.route("/request_control", methods=["POST"])
    @server.restrict
    def request_control():
        """ """

        @copy_current_request_context
        def handle_timeout_gives_control(sid, timeout=30):
            gevent.sleep(timeout)

            if app.TIMEOUT_GIVES_CONTROL:
                # Pass control to user if still waiting
                if current_user.requests_control:
                    toggle_operator(
                        current_user.username, "Timeout expired, you have control"
                    )

        data = request.get_json()

        # Is someone already asking for control
        for observer in app.usermanager.get_observers():
            if observer.requests_control and observer.username != current_user.username:
                msg = "Another user is already asking for control"
                return make_response(msg, 409)

        current_user.requests_control = data["control"]
        server.user_datastore.commit()

        gevent.spawn(handle_timeout_gives_control, current_user.username, timeout=10)

        app.usermanager.emit_observers_changed()

        return make_response("", 200)

    @bp.route("/take_control", methods=["POST"])
    @server.restrict
    def take_control():
        """ """
        # Already master do nothing
        if app.usermanager.is_operator():
            return make_response("", 200)

        # Not inhouse user so not allowed to take control by force,
        # return error code
        if not current_user.isstaff():
            return make_response("", 409)

        toggle_operator(current_user.username, "You were given control")

        return make_response("", 200)

    @bp.route("/give_control", methods=["POST"])
    @server.restrict
    def give_control():
        """ """
        username = request.get_json().get("username")
        toggle_operator(username, "You were given control")

        return make_response("", 200)

    @bp.route("/update_user_nickname", methods=["POST"])
    @server.restrict
    def update_user_nickname():
        """ """
        name = request.get_json().get("name")
        current_user.nickname = name
        app.usermanager.update_user(current_user)
        app.usermanager.emit_observers_changed()

        return make_response("", 200)

    @bp.route("/logout_user", methods=["POST"])
    @server.restrict
    def logout_user():
        """ """
        username = request.get_json().get("username")
        app.usermanager.force_signout_user(username)
        return make_response("", 200)

    def toggle_operator(username, message):
        oldop = app.usermanager.get_operator()
        newop = app.usermanager.set_operator(username)

        join_room("observers", sid=oldop.socketio_session_id, namespace="/ui_state")
        leave_room("observers", sid=newop.socketio_session_id, namespace="/ui_state")

        app.usermanager.emit_observers_changed(message)

    def remain_observer(user, message):
        observer = user.todict()
        observer["message"] = message

        server.emit(
            "setObserver", observer, room=user.socketio_session_id, namespace="/hwr"
        )

    @bp.route("/", methods=["GET"])
    @server.restrict
    def rasettings():
        """ """
        data = {
            "observers": [_u.todict() for _u in app.usermanager.get_observers()],
            "allowRemote": app.ALLOW_REMOTE,
            "timeoutGivesControl": app.TIMEOUT_GIVES_CONTROL,
        }

        return jsonify(data=data)

    @bp.route("/allow_remote", methods=["POST"])
    @server.restrict
    def allow_remote():
        """ """
        allow = request.get_json().get("allow")

        if app.ALLOW_REMOTE and allow == False:
            server.emit("forceSignoutObservers", {}, namespace="/hwr")

        app.ALLOW_REMOTE = allow

        return Response(status=200)

    @bp.route("/timeout_gives_control", methods=["POST"])
    @server.restrict
    def timeout_gives_control():
        """ """
        control = request.get_json().get("timeoutGivesControl")
        app.TIMEOUT_GIVES_CONTROL = control

        return Response(status=200)

    def observer_requesting_control():
        observer = None

        for o in app.usermanager.get_observers():
            if o.requests_control:
                observer = o

        return observer

    @bp.route("/request_control_response", methods=["POST"])
    @server.restrict
    def request_control_response():
        """ """
        data = request.get_json()
        new_op = observer_requesting_control()

        # Request was denied
        if not data["giveControl"]:
            remain_observer(new_op, data["message"])
        else:
            toggle_operator(new_op.username, data["message"])

        new_op.requests_control = False
        app.usermanager.update_user(new_op)

        return make_response("", 200)

    @bp.route("/chat", methods=["POST"])
    def append_message():
        message = request.get_json().get("message", "")
        username = request.get_json().get("username", "")

        if message and username:
            app.chat.append_message(message, username)

        return Response(status=200)

    @bp.route("/chat", methods=["GET"])
    @server.restrict
    def get_all_mesages():
        return jsonify({"messages": app.chat.get_all_messages()})

    @server.flask_socketio.on("connect", namespace="/hwr")
    @server.ws_restrict
    def connect():
        current_user.socketio_session_id = request.sid
        app.usermanager.update_user(current_user)

    @server.flask_socketio.on("disconnect", namespace="/hwr")
    def disconnect():
        if current_user.is_anonymous:
            return
        
        current_user.socketio_session_id = None
        current_user.socketio_session_id = request.sid

    return bp



"""
Module that contains application wide settings and state as well as functions
for accessing and manipulating those.
"""
import os
import sys
import logging
import traceback
import atexit
import json
import time

from pathlib import Path
from urllib.parse import urlparse
from logging import StreamHandler
from logging.handlers import TimedRotatingFileHandler

from mxcubecore import HardwareRepository as HWR
from mxcubecore import removeLoggingHandlers, ColorFormatter
from mxcubecore import queue_entry
from mxcubecore.utils.conversion import make_table

from mxcube3.logging_handler import MX3LoggingHandler
from mxcube3.core.util.adapterutils import get_adapter_cls_from_hardware_object
from mxcube3.core.adapter.adapter_base import AdapterBase
from mxcube3.core.components.component_base import import_component
from mxcube3.core.components.lims import Lims
from mxcube3.core.components.chat import Chat
from mxcube3.core.components.samplechanger import SampleChanger
from mxcube3.core.components.beamline import Beamline
from mxcube3.core.components.sampleview import SampleView
from mxcube3.core.components.queue import Queue
from mxcube3.core.components.workflow import Workflow
from mxcube3.core.components.gphl_workflow import GphlWorkflow


removeLoggingHandlers()


class MXCUBECore:
    # The HardwareRepository object
    hwr = None

    # Below, all the HardwareObjects made available through this module,
    # Initialized by the init function

    # XMLRPCServer
    actions = None
    # Plotting
    plotting = None

    adapter_dict = {}

    @staticmethod
    def exit_with_error(msg):
        """
        Writes the traceback and msg to the log and exits the application

        :param str msg: Additional message to write to log

        """
        logging.getLogger("HWR").error(traceback.format_exc())

        if msg:
            logging.getLogger("HWR").error(msg)

        msg = "Could not initialize one or several hardware objects, stopped "
        msg += "at first error !"

        logging.getLogger("HWR").error(msg)
        logging.getLogger("HWR").error("Quitting server !")
        sys.exit(-1)

    @staticmethod
    def init(app):
        """
        Initializes the HardwareRepository with XML files read from hwdir.

        The hwr module must be imported at the very beginning of the application
        start-up to function correctly.

        This method can however be called later, so that initialization can be
        done when one wishes.

        :param app: FIXME ???

        :return: None
        """
        from mxcube3.core.adapter.beamline_adapter import BeamlineAdapter

        fname = os.path.dirname(__file__)
        HWR.add_hardware_objects_dirs([os.path.join(fname, "HardwareObjects")])
        # rhfogh 20210916. The change allows (me) simpler configuration handling
        # and because of changes in init_hardware_repository does not change
        # current functionality.
        _hwr = HWR.get_hardware_repository()

        MXCUBECore.hwr = _hwr

        try:
            MXCUBECore.beamline = BeamlineAdapter(HWR.beamline, MXCUBEApplication)
            MXCUBECore.adapt_hardware_objects(app)
        except Exception:
            msg = "Could not initialize one or several hardware objects, "
            msg += "stopped at first error ! \n"
            msg += "Make sure That all devices servers are running \n"
            msg += "Make sure that the detector software is running \n"
            MXCUBECore.exit_with_error(msg)

    @staticmethod
    def _get_object_from_id(_id):
        if _id in MXCUBECore.adapter_dict:
            return MXCUBECore.adapter_dict[_id]["adapter"]

    @staticmethod
    def _get_adapter_id(ho):
        _id = HWR.beamline.get_id(ho)

        return _id.replace(" ", "_").lower()

    @staticmethod
    def _add_adapter(_id, adapter_cls, ho, adapter_instance):
        if _id not in MXCUBECore.adapter_dict:
            MXCUBECore.adapter_dict[_id] = {
                "id": str(_id),
                "adapter_cls": adapter_cls.__name__,
                "ho": ho.name()[1:],
                "adapter": adapter_instance,
            }
        else:
            logging.getLogger("MX3.HWR").warning(
                f"Skipping {ho.name()}, id: {_id} already exists"
            )

    @staticmethod
    def get_adapter(_id):
        return MXCUBECore._get_object_from_id(_id)

    @staticmethod
    def adapt_hardware_objects(app):
        adapter_config = app.CONFIG.app.adapter_properties
        hwobject_list = [item for item in MXCUBECore.hwr.hardware_objects]

        for ho_name in hwobject_list:
            # Go through all hardware objects exposed by mxcubecore
            # hardware repository set id to username if its deinfed
            # use the name otherwise (file name without extension)
            ho = MXCUBECore.hwr.get_hardware_object(ho_name)

            if not ho:
                continue

            _id = HWR.beamline.get_id(ho)

            # Try to use the interface exposed by abstract classes in mxcubecore to adapt
            # the object
            adapter_cls = get_adapter_cls_from_hardware_object(ho)

            if adapter_cls:
                try:
                    adapter_instance = adapter_cls(ho, _id, app, **dict(adapter_config))
                    logging.getLogger("MX3.HWR").info("Added adapter for %s" % _id)
                except Exception:
                    logging.getLogger("MX3.HWR").exception(
                        "Could not add adapter for %s" % _id
                    )
                    logging.getLogger("MX3.HWR").info("%s not available" % _id)
                    adapter_cls = AdapterBase
                    adapter_instance = AdapterBase(None, _id, app)

                MXCUBECore._add_adapter(_id, adapter_cls, ho, adapter_instance)
            else:
                logging.getLogger("MX3.HWR").info("No adapter for %s" % _id)

        print(
            make_table(
                ["Beamline attribute (id)", "Adapter", "HO filename"],
                [
                    [item["id"], item["adapter_cls"], item["ho"]]
                    for item in MXCUBECore.adapter_dict.values()
                ],
            )
        )


class MXCUBEApplication:
    t0 = time.time()
    # Below variables used for internal application state

    # SampleID and sample data of currently mounted sample, to handle samples
    # that are not mounted by sample changer.
    CURRENTLY_MOUNTED_SAMPLE = ""

    # Sample location of sample that are in process of being mounted
    SAMPLE_TO_BE_MOUNTED = ""

    # Method used for sample centring
    CENTRING_METHOD = queue_entry.CENTRING_METHOD.LOOP

    # Look up table for finding the limsID for a corresponding queueID (QueueNode)
    NODE_ID_TO_LIMS_ID = {}

    # Initial file list for user, initialized at login, for creating automatic
    # run numbers
    INITIAL_FILE_LIST = []

    # Lookup table for sample changer location to data matrix or
    # data matrix to location
    SC_CONTENTS = {"FROM_CODE": {}, "FROM_LOCATION": {}}

    # Current sample list, with tasks
    SAMPLE_LIST = {"sampleList": {}, "sampleOrder": []}

    # Users currently logged in
    USERS = {}

    # Path to video device (i.e. /dev/videoX)
    VIDEO_FORMAT = "MPEG1"

    # Contains the complete client side ui state, managed up state_storage.py
    UI_STATE = dict()
    TEMP_DISABLED = []

    # Below variables used for application wide settings

    # Enabled or Disable remote usage
    ALLOW_REMOTE = False

    # Enable timeout gives control (if ALLOW_REMOTE is True)
    TIMEOUT_GIVES_CONTROL = False

    # Enable automatic Mountie of sample when queue executed in
    # "automatic/pipeline" mode
    AUTO_MOUNT_SAMPLE = False

    # Automatically add and execute diffraction plans coming from
    # characterizations
    AUTO_ADD_DIFFPLAN = False

    # Number of sample snapshots taken before collect
    NUM_SNAPSHOTS = 4

    CONFIG = None

    mxcubecore = MXCUBECore()

    server = None

    @staticmethod
    def init(
        server, allow_remote, ra_timeout, log_fpath, log_level, enabled_logger_list, cfg
    ):
        """
        Initializes application wide variables, sample video stream, and applies

        :param hwr: HardwareRepository module
        :param bool allow_remote: Allow remote usage, True else False
        :param bool ra_timeout: Timeout gives control, True else False

        :return None:
        """
        logging.getLogger("MX3.HWR").info("Starting MXCuBE3...")
        MXCUBEApplication.server = server
        MXCUBEApplication.ALLOW_REMOTE = allow_remote
        MXCUBEApplication.TIMEOUT_GIVES_CONTROL = ra_timeout
        MXCUBEApplication.CONFIG = cfg

        MXCUBEApplication.mxcubecore.init(MXCUBEApplication)

        if cfg.app.USE_EXTERNAL_STREAMER:
            MXCUBEApplication.init_sample_video(
                _format=cfg.app.VIDEO_FORMAT,
                port=str(urlparse(cfg.app.VIDEO_STREAM_URL).port),
            )

        MXCUBEApplication.init_logging(log_fpath, log_level, enabled_logger_list)

        _UserManagerCls = import_component(
            cfg.app.usermanager, package="components.user"
        )

        MXCUBEApplication.queue = Queue(MXCUBEApplication, {})
        MXCUBEApplication.lims = Lims(MXCUBEApplication, {})
        MXCUBEApplication.usermanager = _UserManagerCls(
            MXCUBEApplication, cfg.app.usermanager
        )
        MXCUBEApplication.chat = Chat(MXCUBEApplication, {})
        MXCUBEApplication.sample_changer = SampleChanger(MXCUBEApplication, {})
        MXCUBEApplication.beamline = Beamline(MXCUBEApplication, {})
        MXCUBEApplication.sample_view = SampleView(MXCUBEApplication, {})
        MXCUBEApplication.workflow = Workflow(MXCUBEApplication, {})
        MXCUBEApplication.gphl_workflow = GphlWorkflow(MXCUBEApplication, {})

        MXCUBEApplication.init_signal_handlers()
        atexit.register(MXCUBEApplication.app_atexit)

        # Install server-side UI state storage
        MXCUBEApplication.init_state_storage()

        # MXCUBEApplication.load_settings()
        msg = "MXCuBE 3 initialized, it took %.1f seconds" % (time.time() - MXCUBEApplication.t0)
        logging.getLogger("MX3.HWR").info(msg)

    @staticmethod
    def init_sample_video(_format, port):
        """
        Initializes video streaming
        :return: None
        """
        try:
            HWR.beamline.sample_view.camera.start_streaming(_format=_format, port=port)
        except Exception as ex:
            msg = "Could not initialize video, error was: "
            msg += str(ex)
            logging.getLogger("HWR").info(msg)

    @staticmethod
    def init_signal_handlers():
        """
        Connects the signal handlers defined in routes/signals.py to the
        corresponding signals/events
        """
        try:
            MXCUBEApplication.queue.init_signals(HWR.beamline.queue_model)
        except Exception:
            sys.excepthook(*sys.exc_info())

        try:
            MXCUBEApplication.sample_view.init_signals()
        except Exception:
            sys.excepthook(*sys.exc_info())

        try:
            MXCUBEApplication.sample_changer.init_signals()
        except Exception:
            sys.excepthook(*sys.exc_info())

        try:
            MXCUBEApplication.beamline.init_signals()
            MXCUBEApplication.beamline.diffractometer_init_signals()
        except Exception:
            sys.excepthook(*sys.exc_info())

    @staticmethod
    def init_logging(log_file, log_level, enabled_logger_list):
        """
        :param str log_file: Path to log file

        :return: None
        """
        removeLoggingHandlers()

        fmt = "%(asctime)s |%(name)-7s|%(levelname)-7s| %(message)s"
        log_formatter = ColorFormatter(fmt)

        if log_file:
            os.chmod(log_file, 0o666)
            Path(log_file).touch()

            log_file_handler = TimedRotatingFileHandler(
                log_file, when="midnight", backupCount=7
            )
            log_file_handler.setFormatter(log_formatter)

            uilog_file = f"{log_file[:-4]}_ui.log"
            os.chmod(uilog_file, 0o666)
            Path(uilog_file).touch()

            uilog_file_handler = TimedRotatingFileHandler(
                uilog_file, when="midnight", backupCount=7
            )
            uilog_file_handler.setFormatter(log_formatter)

        if not log_level:
            log_level = "INFO"
        else:
            log_level = log_level.upper()

        root_logger = logging.getLogger()
        root_logger.setLevel(log_level)

        custom_log_handler = MX3LoggingHandler(MXCUBEApplication.server)
        custom_log_handler.setLevel(log_level)
        custom_log_handler.setFormatter(log_formatter)

        _loggers = {
            "exception_logger": logging.getLogger("exceptions"),
            "hwr_logger": logging.getLogger("HWR"),
            "mx3_hwr_logger": logging.getLogger("MX3.HWR"),
            "user_logger": logging.getLogger("user_level_log"),
            "queue_logger": logging.getLogger("queue_exec"),
            "mx3_ui_logger": logging.getLogger("MX3.UI"),
        }

        stdout_log_handler = StreamHandler(sys.stdout)
        stdout_log_handler.setFormatter(log_formatter)

        for logger_name, logger in _loggers.items():
            if logger_name in enabled_logger_list:
                logger.addHandler(custom_log_handler)
                logger.addHandler(stdout_log_handler)
                logger.setLevel(log_level)

                if log_file and "mx3_ui" in logger_name:
                    logger.addHandler(uilog_file_handler)
                elif log_file:
                    logger.addHandler(log_file_handler)

                logger.propagate = False
            else:
                logger.disabled = True

    @staticmethod
    def init_state_storage():
        """
        Set up of server side state storage, the UI state of the client is
        stored on the server
        """
        from mxcube3 import state_storage

        state_storage.init()

    @staticmethod
    def get_ui_properties():
        # Add type information to each component retrieved from the beamline adapter
        # (either via config or via mxcubecore.beamline)
        for _item_name, item_data in MXCUBEApplication.CONFIG.app.ui_properties.items():
            for component_data in item_data.components:
                try:
                    mxcore = MXCUBEApplication.mxcubecore
                    adapter = mxcore.get_adapter(component_data.attribute)
                    adapter_cls_name = type(adapter).__name__
                    value_type = adapter.adapter_type
                except AttributeError:
                    msg = f"{component_data.attribute} not accessible via Beamline object. "
                    msg += f"Verify that beamline.{component_data.attribute} is valid and/or "
                    msg += f"{component_data.attribute} accessible via get_role "
                    msg += "check ui.yaml configuration file. "
                    msg += "(attribute will NOT be avilable in UI)"
                    logging.getLogger("HWR").warning(msg)
                    adapter_cls_name = ""
                    value_type = ""
                else:
                    adapter_cls_name = adapter_cls_name.replace("Adapter", "")

                if not component_data.object_type:
                    component_data.object_type = adapter_cls_name

                if not component_data.value_type:
                    component_data.value_type = value_type

        return {
            key: value.dict()
            for (key, value) in MXCUBEApplication.CONFIG.app.ui_properties.items()
        }

    @staticmethod
    def save_settings():
        """
        Saves all application wide variables to disk, stored-mxcube-session.json
        """
        queue = MXCUBEApplication.queue.queue_to_dict(
            HWR.beamline.queue_model.get_model_root()
        )

        # For the moment not storing USERS

        data = {
            "QUEUE": queue,
            "CURRENTLY_MOUNTED_SAMPLE": MXCUBEApplication.CURRENTLY_MOUNTED_SAMPLE,
            "SAMPLE_TO_BE_MOUNTED": MXCUBEApplication.SAMPLE_TO_BE_MOUNTED,
            "CENTRING_METHOD": MXCUBEApplication.CENTRING_METHOD,
            "NODE_ID_TO_LIMS_ID": MXCUBEApplication.NODE_ID_TO_LIMS_ID,
            "INITIAL_FILE_LIST": MXCUBEApplication.INITIAL_FILE_LIST,
            "SC_CONTENTS": MXCUBEApplication.SC_CONTENTS,
            "SAMPLE_LIST": MXCUBEApplication.SAMPLE_LIST,
            "TEMP_DISABLED": MXCUBEApplication.TEMP_DISABLED,
            "ALLOW_REMOTE": MXCUBEApplication.ALLOW_REMOTE,
            "TIMEOUT_GIVES_CONTROL": MXCUBEApplication.TIMEOUT_GIVES_CONTROL,
            "VIDEO_FORMAT": MXCUBEApplication.VIDEO_FORMAT,
            "AUTO_MOUNT_SAMPLE": MXCUBEApplication.AUTO_MOUNT_SAMPLE,
            "AUTO_ADD_DIFFPLAN": MXCUBEApplication.AUTO_ADD_DIFFPLAN,
            "NUM_SNAPSHOTS": MXCUBEApplication.NUM_SNAPSHOTS,
            "UI_STATE": MXCUBEApplication.UI_STATE,
        }

        fname = Path("/tmp/stored-mxcube-session.json")
        fname.touch(exist_ok=True)

        with open(fname, "w+") as fp:
            json.dump(data, fp)

    @staticmethod
    def load_settings():
        """
        Loads application wide variables from "stored-mxcube-session.json"
        """
        with open("/tmp/stored-mxcube-session.json", "r") as f:
            data = json.load(f)

        MXCUBEApplication.queue.load_queue_from_dict(data.get("QUEUE", {}))

        MXCUBEApplication.CENTRING_METHOD = data.get(
            "CENTRING_METHOD", queue_entry.CENTRING_METHOD.LOOP
        )
        MXCUBEApplication.NODE_ID_TO_LIMS_ID = data.get("NODE_ID_TO_LIMS_ID", {})
        MXCUBEApplication.SC_CONTENTS = data.get(
            "SC_CONTENTS", {"FROM_CODE": {}, "FROM_LOCATION": {}}
        )
        MXCUBEApplication.SAMPLE_LIST = data.get(
            "SAMPLE_LIST", {"sampleList": {}, "sampleOrder": []}
        )
        MXCUBEApplication.ALLOW_REMOTE = data.get("ALLOW_REMOTE", False)
        MXCUBEApplication.TIMEOUT_GIVES_CONTROL = data.get(
            "TIMEOUT_GIVES_CONTROL", False
        )
        MXCUBEApplication.AUTO_MOUNT_SAMPLE = data.get("AUTO_MOUNT_SAMPLE", False)
        MXCUBEApplication.AUTO_ADD_DIFFPLAN = data.get("AUTO_ADD_DIFFPLAN", False)
        MXCUBEApplication.NUM_SNAPSHOTS = data.get("NUM_SNAPSHOTS", False)
        MXCUBEApplication.UI_STATE = data.get("UI_STATE", {})

    @staticmethod
    def app_atexit():
        MXCUBEApplication.save_settings()



# -*- coding: utf-8 -*-
import io

from flask import Blueprint, Response, jsonify, request, send_file


def init_route(app, server, url_prefix):
    bp = Blueprint("gphl_workflow", __name__, url_prefix=url_prefix)

    @bp.route("/", methods=["GET"])
    @server.restrict
    def gphlworkflow():
        return jsonify(app.gphl_workflow.get_available_workflows())

    @bp.route("/", methods=["POST"])
    @server.restrict
    def submit_parameters():
        data = request.get_json()
        app.gphl_workflow.submit_parameters(data)
        return Response(status=200)

    return bp



from flask import Blueprint, jsonify
from mxcube3.core.models.generic import PathModel, Response


def init_route(app, server, url_prefix):
    bp = Blueprint("detector", __name__, url_prefix=url_prefix)

    @bp.route("/", methods=["GET"])
    @server.restrict
    def get_detector_info():
        """
        Retrieves general info from the detector.
            :response Content-type: application/json, example:
                {'filetype': 'h5'},
            :statuscode: 200: no error
            :statuscode: 409: error
        """

        resp = jsonify({"fileSuffix": app.beamline.get_detector_info()})
        resp.status_code = 200
        return resp

    @bp.route("/display_image/", methods=["GET"])
    @server.restrict
    @server.validate(resp=Response(HTTP_200=""))
    def display_image(query: PathModel):
        return ""

    return bp



import os
import json

from flask import Blueprint, Response, jsonify, request

from mxcubecore import HardwareRepository as HWR


def init_route(app, server, url_prefix):
    bp = Blueprint("sampleview", __name__, url_prefix=url_prefix)

    @bp.route("/camera/subscribe", methods=["GET"])
    @server.restrict
    def subscribe_to_camera():
        """
        Subscribe to the camera streaming
            :response: image as html Content-type
        """
        if app.CONFIG.app.VIDEO_FORMAT == "MPEG1":
            result = Response(status=200)
        else:
            result = app.sample_view.http_streamer.get_response()

        return result

    @bp.route("/camera/unsubscribe", methods=["PUT"])
    @server.restrict
    def unsubscribe_to_camera():
        """
        SampleCentring: unsubscribe from the camera streaming
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        HWR.beamline.sample_view.camera.streaming_greenlet.kill()
        return Response(status=200)

    @bp.route("/camera/save", methods=["PUT"])
    @server.restrict
    def snapshot():
        """
        Save snapshot of the sample view
        data = {generic_data, "Path": path} # not sure if path should be available,
        or directly use the user/proposal path
        Return: 'True' if command issued succesfully, otherwise 'False'.
        """
        try:
            HWR.beamline.sample_view.camera.takeSnapshot(
                os.path.join(os.path.dirname(__file__), "snapshots/")
            )
            return "True"
        except Exception:
            return "False"

    @bp.route("/camera", methods=["GET"])
    @server.restrict
    def get_image_data():
        """
        Get size of the image of the diffractometer
            :response Content-type:application/json, example:
            {  "imageHeight": 576, "imageWidth": 768,
            "pixelsPerMm": [1661.1295681063123, 1661.1295681063123]
            }
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        data = app.beamline.get_viewport_info()

        resp = jsonify(data)
        resp.status_code = 200
        return resp

    @bp.route("/camera", methods=["POST"])
    @server.restrict
    def set_image_size():
        """ """
        params = request.get_json()

        res = app.sample_view.set_image_size(
            float(params["width"]), float(params["height"])
        )

        resp = jsonify(res)
        resp.status_code = 200
        return resp

    @bp.route("/centring/<point_id>/moveto", methods=["PUT"])
    @server.require_control
    @server.restrict
    def move_to_centred_position(point_id):
        """
        Move to the given centred position.
            :parameter id: centred position identifier, integer
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        point = app.sample_view.move_to_centred_position(point_id)

        if point:
            return Response(status=200)
        else:
            return Response(status=409)

    @bp.route("/shapes", methods=["GET"])
    @server.restrict
    def get_shapes():
        """
        Retrieve all the stored centred positions.
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        shapes = app.sample_view.get_shapes()

        resp = jsonify(shapes)
        resp.status_code = 200
        return resp

    @bp.route("/shapes/<sid>", methods=["GET"])
    @server.restrict
    def get_shape_with_sid(sid):
        """
        Retrieve requested shape information.
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: shape not found
        """
        shape = app.sample_view.get_shapes()

        if shape is not None:
            resp = jsonify(shape)
            resp.status_code = 200
            return resp
        else:
            return Response(status=409)

    @bp.route("/shapes/<sid>", methods=["POST"])
    def shape_add_cell_result(sid):
        """
        Update cell result data.
            :parameter shape_data: dict with result info (cell number, result value)
            :response Content-type: application/json, response status.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        params = request.get_json()

        cell_number = params.get("cell", 0)
        result = params.get("result", 0)

        app.sample_view.shape_add_cell_result(sid, cell_number, result)
        return Response(status=200)

    @bp.route("/shapes", methods=["POST"])
    @server.require_control
    @server.restrict
    def update_shapes():
        """
        Update shape information.
            :parameter shape_data: dict with shape information (id, type, ...)
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        shapes = request.get_json().get("shapes", [])

        resp = jsonify(app.sample_view.update_shapes(shapes))
        resp.status_code = 200

        return resp

    @bp.route("/shapes/<sid>", methods=["DELETE"])
    @server.require_control
    @server.restrict
    def delete_shape(sid):
        """
        Retrieve all the stored centred positions.
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        HWR.beamline.sample_view.delete_shape(sid)
        return Response(status=200)

    @bp.route("/shapes/rotate_to", methods=["POST"])
    @server.require_control
    @server.restrict
    def rotate_to():
        """
        Rotate Phi to the position where the given shape was defined

            :parameter sid: The shape id
            :response Content-type: application/json, the stored centred positions.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        sid = request.get_json().get("sid", -1)

        try:
            app.sample_view.rotate_to(sid)
        except Exception:
            resp = Response(status=409)
        else:
            resp = Response(status=200)

        return resp

    @bp.route("/zoom", methods=["PUT"])
    @server.require_control
    @server.restrict
    def move_zoom_motor():
        """
        Move the zoom motor.
            :request Content-type: application/json, new position {'level': 4}.
            Note: level specified as integer (not 'Zoom 4')
            :response Content-type: application/json, new scale value,
            example: {"pixelsPerMm": [ 1661.1, 1661.1]}
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        params = request.data
        pos = json.loads(params).get("level", 0)

        res = app.sample_view.move_zoom_motor(pos)

        resp = jsonify(res)
        resp.status_code = 200
        return resp

    @bp.route("/backlighton", methods=["PUT"])
    @server.require_control
    @server.restrict
    def back_light_on():
        """
        Activate the backlight of the diffractometer.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        app.sample_view.back_light_on()
        return Response(status=200)

    @bp.route("/backlightoff", methods=["PUT"])
    @server.require_control
    @server.restrict
    def back_light_off():
        """
        Switch off the backlight of the diffractometer.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        app.sample_view.back_light_off()
        return Response(status=200)

    @bp.route("/frontlighton", methods=["PUT"])
    @server.require_control
    @server.restrict
    def front_light_on():
        """
        Activate the frontlight of the diffractometer.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        app.sample_view.front_light_on()
        return Response(status=200)

    @bp.route("/frontlightoff", methods=["PUT"])
    @server.require_control
    @server.restrict
    def front_light_off():
        """
        Switch off the frontlight of the diffractometer.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        app.sample_view.front_light_off()
        return Response(status=200)

    @bp.route("/<motid>/<newpos>", methods=["PUT"])
    @server.require_control
    @server.restrict
    def move_motor(motid, newpos):
        """
        Move or Stop the given motor.
        :parameter motid: motor name, 'Phi', 'Focus', 'PhiZ', 'PhiY', 'Zoom',
        'BackLightSwitch','BackLight','FrontLightSwitch', 'FrontLight',
        'Sampx', 'Sampy'
        :parameter newpos: new position, double, stop: string
        :statuscode: 200: no error
        :statuscode: 409: error
        """

        try:
            app.sample_view.move_motor(motid, newpos)
        except Exception as ex:
            return (
                "Could not move motor %s" % str(ex),
                409,
                {"Content-Type": "application/json", "msg": str(ex)},
            )
        else:
            return Response(status=200)

    # @bp.route("/<elem_id>", methods=["GET"])
    # @server.restrict
    # def get_status_of_id(elem_id):
    #     """
    #     Get position and status of the given element
    #         :parameter id: moveable to get its status, 'Phi', 'Focus', 'PhiZ',
    #         'PhiY', 'Zoom', 'BackLightSwitch','BackLight','FrontLightSwitch',
    #         'FrontLight','Sampx', 'Sampy'
    #         :response Content-type: application/json, {motorname:
    #             {'Status': status, 'position': position} }
    #         :statuscode: 200: no error
    #         :statuscode: 409: error
    #     """

    #     try:
    #         ret = app.sample_view.get_status_of_id(elem_id)
    #         resp = jsonify(ret)
    #         resp.status_code = 200
    #         return resp
    #     except Exception:
    #         return Response(status=409)

    @bp.route("/centring/startauto", methods=["GET"])
    @server.require_control
    @server.restrict
    def centre_auto():
        """
        Start automatic (lucid) centring procedure.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        app.sample_view.start_auto_centring()
        return Response(status=200)

    @bp.route("/centring/start3click", methods=["PUT"])
    @server.require_control
    @server.restrict
    def centre_3_click():
        """
        Start 3 click centring procedure.
            :statuscode: 200: no error
            :statuscode: 409: error
        """

        try:
            data = app.sample_view.start_manual_centring()
        except Exception as ex:
            resp = (
                "Could not move motor %s" % str(ex),
                409,
                {"Content-Type": "application/json", "msg": str(ex)},
            )
        else:
            resp = jsonify(data)
            resp.status_code = 200

        return resp

    @bp.route("/centring/abort", methods=["PUT"])
    @server.require_control
    @server.restrict
    def abort_centring():
        """
        Abort centring procedure.
            :statuscode: 200: no error
            :statuscode: 409: error
        """
        app.sample_view.abort_centring()
        return Response(status=200)

    @bp.route("/centring/click", methods=["PUT"])
    @server.require_control
    @server.restrict
    def click():
        """
        The 3-click method needs the input from the user, a running 3click centring
        procedure must be set before

        :request Content-type: application/json, integer positions of the clicks,
                            {clickPos={"x": 123,"y": 456}}
        :response Content-type: application/json, integer, number of clicks
                                left {'clickLeft': 3 | 2 | 1}
        :statuscode: 200: no error
        :statuscode: 409: error
        """
        pos = json.loads(request.data).get("clickPos", None)

        data = app.sample_view.centring_handle_click(pos["x"], pos["y"])

        resp = jsonify(data)
        resp.status_code = 200
        return resp

    @bp.route("/centring/accept", methods=["PUT"])
    @server.require_control
    @server.restrict
    def accept_centring():
        """
        Accept the centring position.
        """
        HWR.beamline.diffractometer.accept_centring()
        return Response(status=200)

    @bp.route("/centring/reject", methods=["PUT"])
    @server.require_control
    @server.restrict
    def reject_centring():
        """Reject the centring position."""
        app.sample_view.reject_centring()
        return Response(status=200)

    @bp.route("/movetobeam", methods=["PUT"])
    @server.require_control
    @server.restrict
    def move_to_beam():
        """Go to the beam position from the given (x, y) position."""
        pos = json.loads(request.data).get("clickPos")

        app.sample_view.move_to_beam(pos["x"], pos["y"])

        return Response(status=200)

    @bp.route("/centring/centring_method", methods=["PUT"])
    @server.require_control
    @server.restrict
    def set_centring_method():
        """
        Set MXCuBE to use automatic (lucid) centring procedure when
        mounting samples

        :statuscode: 200: no error
        :statuscode: 409: error

        """
        method = json.loads(request.data).get("centringMethod", None)
        app.sample_view.set_centring_method(method)
        return Response(status=200)

    return bp
